from pathlib import Path, PurePath, _PosixFlavour
from typing import Callable
import logging
import functools
import abc
from contextlib import contextmanager
import re
import os

logger = logging.getLogger(__name__)


class _URIFlavour(_PosixFlavour):
    has_drv = True
    is_supported = True

    def splitroot(self, part, sep=_PosixFlavour.sep):
        assert sep == self.sep

        match = re.match(rf"(.*):{re.escape(sep)}{{2}}(.*)", part)
        if match:
            drive, path = match.groups()
            drive = drive + "://"
            root = ""
            return drive, root, path
        else:
            return super().splitroot(part, sep=sep)


class _URIBase(Path):
    _flavour = _URIFlavour()

    @property
    def scheme(self) -> str:
        if not self.drive:
            return ""
        return self.drive[:-len("://")]

    @property
    def path(self) -> str:
        begin = 1 if self.drive or self.root else 0
        return self.root + self._flavour.join(self.parts[begin:])

    def __repr__(self) -> str:
        s = ""
        if self.scheme:
            s += self.scheme + ":" + self._flavour.sep * 2
        s += self.path
        return "{}({!r})".format(self.__class__.__name__, s)


class PathManagerBase:
    """Base class for a path manager.

    This class simultaneously acts as a context manager for the currently active path manager of the URI class.
    """

    def __init__(self):
        self._handlers = {}
        self._previous_path_managers = []

    def resolve(self, path: os.PathLike) -> Path:
        """Resolve a path (which might be a URI) to a local path.

        Args:
            path (os.PathLike): the path

        Returns:
            Path: the corresponding local path
        """

        if not isinstance(path, _URIBase):
            path = _URIBase(path)
        if path.scheme:
            if path.scheme not in self._handlers:
                raise NotImplementedError(f"No handler is registered for scheme {path.scheme}")
            return self._handlers[path.scheme](path)
        else:
            return Path(path.path)

    def register_handler(self, scheme: str) -> callable:
        """Decorator to register a path handler for a given URI scheme.

        Args:
            scheme (str): the scheme

        Returns:
            callable: the decorated function
        """

        def decorator(func: Callable[[os.PathLike], Path]):
            self._handlers[scheme] = func
            logger.debug(f"Registered path handler for scheme {scheme}")
            return func
        return decorator

    def __enter__(self):
        self._previous_path_managers.append(PathManager._instance)
        PathManager._instance = self

    def __exit__(self, exc_type, exc_val, exc_tb):
        PathManager._instance = self._previous_path_managers.pop()


class PathManagerProxy(PathManagerBase):
    """Proxy class for the main path manager instance.
    """

    def __init__(self, instance: PathManagerBase):
        self._instance = instance

    def resolve(self, *args, **kwargs):
        return self._instance.resolve(*args, **kwargs)

    def register_handler(self, *args, **kwargs):
        return self._instance.register_handler(*args, **kwargs)

    def __enter__(self, *args, **kwargs):
        return self._instance.__enter__(*args, **kwargs)

    def __exit__(self, *args, **kwargs):
         return self._instance.__exit__(*args, **kwargs)

#: The public path manager instance.
PathManager: PathManagerBase = PathManagerProxy(PathManagerBase())


class URI(_URIBase):
    """A class representing a recap URI that is lazily evaluated to a local path when it is used.

    It is fully compatible with pathlib.Path.
    """

    def _init(self, *args, **kwargs):
        self._local_path = PathManager.resolve(self)
        super()._init()

    def __str__(self) -> str:
        return str(self._local_path)


class PathTranslator(abc.ABC):
    """Abstract class representing a path translator that can translate a specific type of URIs to local paths.
    """

    def __call__(self, uri: URI) -> Path:
        """Translate a URI to a local path.

        Usually, this involves looking at uri.path.

        Args:
            uri (URI): the URI

        Returns:
            Path: the corresponding local path
        """

        raise NotImplementedError


def register_translator(scheme: str, path: Path):
    """Convenience method to register a path translator that forwards a URI scheme to a local path.

    Args:
        scheme (str): the URI scheme
        path (Path): the local path
    """

    class Translator(PathTranslator):
        def __call__(self, uri: URI) -> Path:
            return path / uri.path
    PathManager.register_handler(scheme)(Translator())



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['recap']

package_data = \
{'': ['*']}

install_requires = \
['pyyaml>=5.3.1,<6.0.0', 'yacs>=0.1.8,<0.2.0']

extras_require = \
{'docs': ['sphinx>=3.4.1,<4.0.0',
          'sphinx-rtd-theme>=0.5.0,<0.6.0',
          'm2r2>=0.2.7,<0.3.0']}

setup_kwargs = {
    'name': 'recap',
    'version': '0.1.4',
    'description': 'Reproducible configurations for any project',
    'long_description': '# recap\n\n![build](https://github.com/georgw777/recap/workflows/build/badge.svg)\n![PyPI](https://img.shields.io/pypi/v/recap)\n![PyPI - Python Version](https://img.shields.io/pypi/pyversions/recap)\n![GitHub](https://img.shields.io/github/license/georgw777/recap)\n\n_recap_ is a tool for providing _REproducible Configurations for Any Project_.\n\nResearch should be reproducible.\nEspecially in deep learning, it is important to keep track of hyperparameters and configurations used in experiments.\nThis package aims at making that easier.\n\n## Installing\n\nJust install like any Python package:\n\n```bash\npip install recap\n```\n\n## Overview\n\nRecap provides two top-level concepts that would be imported as follows:\n\n```python\nfrom recap import URI, CfgNode as CN\n```\n\nThe `CfgNode` is a subclass of [yacs](https://github.com/rbgirshick/yacs)\' `CfgNode`.\nIt provides some additional features for parsing configurations that are inherited between files which is not possible with yacs.\n\nRecap\'s `URI` class provides a mechanism for handling logical paths within your project more conveniently with an interface that is fully compatible with `pathlib.Path`.\n\n## YAML configurations\n\nConfigurations are defined [just like in yacs](https://github.com/rbgirshick/yacs#usage), except that you need to import the `CfgNode` class from the recap package instead of yacs.\nConsider the following YAML configuration that sets default values for all configuration options we will use in our project. We shall name it `_base.yaml` because our experiments will build on these values.\n\n```yaml\nSYSTEM:\n  NUM_GPUS: 4\n  NUM_WORKERS: 2\nTRAIN:\n  LEARNING_RATE: 0.001\n  BATCH_SIZE: 32\n  SOME_OTHER_HYPERPARAMETER: 10\n```\n\nThe equivalent configuration can be obtained programatically like so:\n\n```python\nfrom recap import CfgNode as CN\n\ncfg = CN()\ncfg.SYSTEM = CN()\ncfg.SYSTEM.NUM_GPUS = 4\ncfg.SYSTEM.NUM_WORKERS = 2\ncfg.TRAIN = CN()\ncfg.TRAIN.LEARNING_RATE = 1e-3\ncfg.TRAIN.BATCH_SIZE = 32\ncfg.TRAIN.SOME_OTHER_HYPERPARAMETER = 10\n\nprint(cfg)\n```\n\n### Inheriting configurations\n\nRecap provides functionality for inheriting configuration options from other configuration files by setting the top-level `_BASE_` key.\nSo, we could create a configuration file `experiment_1.yaml` for an experiment where we try a different learning rate and batch size:\n\n```yaml\n_BASE_: _base.yaml\n\nTRAIN:\n  LEARNING_RATE: 1e-2\n  BATCH_SIZE: 64\n```\n\nIn our code, when we want to load the experiment configuration, we would use the `recap.CfgNode.load_yaml_with_base()` function:\n\n```python\nfrom recap import CfgNode as CN\n\ncfg = CN.load_yaml_with_base("experiment_1.yaml")\n\nprint(cfg)\n\n# Will output:\n"""\nSYSTEM:\n  NUM_GPUS: 4\n  NUM_WORKERS: 2\nTRAIN:\n  LEARNING_RATE: 0.01\n  BATCH_SIZE: 64\n  SOME_OTHER_HYPERPARAMETER: 10\n"""\n```\n\nNote that the `_BASE_` keys can be arbitrarily nested; however, circular references are prohibited.\n\n## Logical URIs and the path manager\n\nRecap includes a path manager for conveniently specifying paths to logical entities.\nThe path strings are set up like a URI where the scheme (i.e. `http` in the path string `http://google.com`) refers to a logical entity.\nEach such entity needs to be set up as a `PathTranslator` that can translate the logical URI path to a physical path on the file system.\n\nFor example, we could set up a path translator for the `data` scheme to refer to the the path of a dataset on our file system located at `/path/to/dataset`. Then the recap URI `data://train/abc.txt` would be translated to `/path/to/dataset/train/abc.txt`.\n\nThe simplest way of setting that up is using the `register_translator` function (although more complex setups are possible with the `recap.path_manager.PathTranslator` class, allowing you to download files from the internet, for example):\n\n```python\nfrom recap.path_manager import register_translator\nfrom pathlib import Path\n\nregister_translator("data", Path("/path/to/dataset"))\n```\n\nThen, we can use the `recap.URI` class just like any `pathlib.Path` object:\n\n```python\nfrom recap import URI\n\nmy_uri = URI("data://train/abc.txt")\n# Here, str(my_uri) == "/path/to/dataset/train/abc.txt"\n\nwith my_uri.open("r") as f:\n    print(f.read())\n```\n\n### Logical URIs in inherited configurations\n\nThe `recap.URI` interface is fully compatible with the nested configurations.\nThis means that you can use recap `URI`s within the `_BASE_` field for inheriting configurations.\n\nFor example, you could register a path translator for the `config` scheme and then include `_BASE_: config://_base.yaml` in your configuration files.\n',
    'author': 'Georg Wölflein',
    'author_email': 'georgw7777@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/georgw777/recap',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
